#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-interface/ampel/util/serialize.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                25.03.2021
# Last Modified Date:  20.12.2021
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

from typing import Any
from base64 import b64encode, b64decode

try:
	from bson import ObjectId
	HAVE_BSON = True
except ImportError:
	HAVE_BSON = False
 
dsi = dict.__setitem__

def walk_and_encode(arg: Any, destructive: bool = True) -> Any:
	"""
	Convert bytes values into "b64:<string>"
	Convert ObjectId instances into "oid:<hex>"
	:param destructive: true modifies the input dict(s)
	"""

	if isinstance(arg, (list, tuple, set)):
		return [walk_and_encode(el, destructive) for el in arg]

	elif isinstance(arg, dict):
		if destructive:
			for k, v in arg.items():
				if HAVE_BSON and isinstance(v, ObjectId):
					dsi(arg, k, "oid:" + v.binary.hex())
				elif isinstance(v, bytes):
					dsi(arg, k, "b64:" + b64encode(v).decode('ascii'))
				else:
					dsi(arg, k, walk_and_encode(v, destructive))
			return arg
		else:
			return {
				k: ("oid:" + v.binary.hex())
				if HAVE_BSON and isinstance(v, ObjectId) else (
					("b64:" + b64encode(v).decode('ascii'))
					if isinstance(v, bytes)
					else walk_and_encode(v, destructive)
				)
				for k, v in arg.items()
			}

	elif HAVE_BSON and isinstance(arg, ObjectId):
		return "oid:" + arg.binary.hex()

	elif isinstance(arg, bytes):
		return "b64:" + b64encode(arg).decode('ascii')

	return arg


def walk_and_decode(arg: Any) -> Any:
	"""
	Convert str values starting with "base64:" into bytes
	Note: modifies the input dict(s)
	"""

	if isinstance(arg, list):
		return [walk_and_decode(el) for el in arg]

	elif isinstance(arg, dict):
		for k, v in arg.items():
			if isinstance(v, str) and v.startswith("b64:"):
				arg[k] = b64decode(v[4:])
			elif isinstance(v, str) and v.startswith("oid:"):
				if HAVE_BSON:
					arg[k] = ObjectId(v[4:])
				else:
					raise ValueError(f"got ObjectId {v}, but pymongo is not installed")

			else:
				arg[k] = walk_and_decode(v)
		return arg

	elif isinstance(arg, str) and arg.startswith("b64:"):
		return b64decode(arg[4:])

	elif isinstance(arg, str) and arg.startswith("oid:"):
		if HAVE_BSON:
			return ObjectId(arg[4:])
		else:
			raise ValueError(f"got ObjectId {arg}, but pymongo is not installed")

	return arg



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-interface/ampel/abstract/AbsTiedPointT2Unit.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                16.02.2021
# Last Modified Date:  28.09.2021
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

from typing import Union
from collections.abc import Sequence
from ampel.types import UBson
from ampel.struct.UnitResult import UnitResult
from ampel.view.T2DocView import T2DocView
from ampel.content.DataPoint import DataPoint
from ampel.base.decorator import abstractmethod
from ampel.abstract.AbsTiedT2Unit import AbsTiedT2Unit


class AbsTiedPointT2Unit(AbsTiedT2Unit, abstract=True):
	"""
	A T2 unit bound to a :class:`~ampel.content.DataPoint.DataPoint` as well
	as the results of other T2 units

	Note that the implementing class can customize the default ingestion behavior
	by defining the class variable 'eligible'.
	See AbsPointT2Unit and DPSelection docstrings for more info
	"""

	@abstractmethod
	def process(self, datapoint: DataPoint, t2_views: Sequence[T2DocView]) -> UBson | UnitResult:
		"""
		Returned object should contain computed science results to be saved into the DB.

		.. note:: the returned dict must have only string keys and be BSON-encodable
		"""



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-interface/ampel/content/MetaActivity.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                07.10.2021
# Last Modified Date:  10.10.2021
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

from typing import TypedDict
from collections.abc import Sequence
from ampel.types import ChannelId, Tag, DataPointId
from ampel.enum.MetaActionCode import MetaActionCode


class MetaActivity(TypedDict, total=False):
	""" A record of activity on tier documents """

	#: Action code(s)
	action: MetaActionCode

	#: Free-form labels
	tag: Tag | Sequence[Tag]

	#: references dps exclusion (t1)
	excl: list[DataPointId]

	#: Document code (useful when t1_compute is performed on the fly)
	code: int

	#: Channel(s) associated with action
	channel: ChannelId | Sequence[ChannelId]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-interface/ampel/protocol/AmpelAlertProtocol.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                24.11.2021
# Last Modified Date:  24.11.2021
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

from typing import Any, Protocol, runtime_checkable
from collections.abc import Sequence
from ampel.types import StockId, Tag

@runtime_checkable
class AmpelAlertProtocol(Protocol):

	@property
	def id(self) -> int:
		...

	@property
	def stock(self) -> StockId:
		...

	@property
	def tag(self) -> None | Tag | list[Tag]:
		...

	@property
	def extra(self) -> None | dict[str, Any]:
		...

	@property
	def datapoints(self) -> Sequence[dict[str, Any]]:
		...

	def get_values(self,
		key: str, filters: None | Sequence[dict[str, Any]] = None
	) -> list[Any]:
		...

	def get_tuples(self,
		key1: str, key2: str,
		filters: None | Sequence[dict[str, Any]] = None
	) -> list[tuple[Any, Any]]:
		...

	def get_ntuples(self,
		params: list[str], filters: None | Sequence[dict[str, Any]] = None
	) -> list[tuple]:
		...

	# Unsure whether this belongs here
	def is_new(self) -> bool:
		...

	def dict(self) -> dict[str, Any]:
		...



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-interface/ampel/base/AuxUnitRegister.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                17.02.2021
# Last Modified Date:  08.11.2021
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

from importlib import import_module
from typing import Any, ClassVar, overload # type: ignore[attr-defined]
from ampel.types import T, check_class
from ampel.model.UnitModel import UnitModel
from ampel.base.AmpelUnit import AmpelUnit
from ampel.config.AmpelConfig import AmpelConfig


class AuxUnitRegister:

	# References unit definitions of auxiliary units
	# allows units to load aux units by themselve or
	# aux units to load other aux units
	_defs: ClassVar[dict[str, Any]] = {}
	_dyn: ClassVar[dict[str, Any]] = {}

	@classmethod
	def initialize(cls, config: AmpelConfig) -> None:
		cls._defs = {
			k: v for k, v in config.get("unit", ret_type=dict, raise_exc=True).items()
			if 'ContextUnit' not in v.get('base', []) and 'LogicalUnit' not in v.get('base', [])
		}

	@overload
	@classmethod
	def new_unit(cls, model: UnitModel, *, sub_type: type[T], **kwargs) -> T:
		...
	@overload
	@classmethod
	def new_unit(cls, model: UnitModel, *, sub_type: None = ..., **kwargs) -> AmpelUnit:
		...

	@classmethod
	def new_unit(cls, model: UnitModel, *, sub_type: None | type[T] = None, **kwargs) -> T | AmpelUnit:
		"""	:raises: ValueError is model.config is not of type None | dict """

		if model.unit in cls._dyn:
			Klass = cls._dyn[model.unit]
		else:
			Klass = cls.get_aux_class(klass=model.unit, sub_type=sub_type)

		if model.config:
			if isinstance(model.config, dict):
				return Klass(**(model.config | kwargs)) # type: ignore[call-arg]
			raise ValueError("Auxiliary units cannot use config aliases")

		unit = Klass(**kwargs) # type: ignore[call-arg]
		if hasattr(unit, "post_init"):
			unit.post_init() # type: ignore[union-attr]

		return unit


	@overload
	@classmethod
	def get_aux_class(cls, klass: str, *, sub_type: type[T]) -> type[T]:
		...

	@overload
	@classmethod
	def get_aux_class(cls, klass: str, *, sub_type: None = ...) -> type[AmpelUnit]:
		...

	@classmethod
	def get_aux_class(cls, klass: str, *, sub_type: None | type[T] = None) -> type[T | type[AmpelUnit]]:
		""" :raises: ValueError if unit is unknown """

		if klass not in cls._defs:
			if not cls._defs:
				raise ValueError(
					f"Unknown auxiliary unit {klass}:"
					f"- make sure a context is loaded (requirement)\n"
				)
			else:
				raise ValueError(
					f"Unknown auxiliary unit {klass}:\n"
					f"- check your ampel conf to see if the unit is properly registered"
				)

		fqn = cls._defs[klass]['fqn']
		ret = getattr(import_module(fqn), klass)

		if sub_type:
			check_class(ret, sub_type)

		return ret



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-interface/ampel/base/AmpelBaseModel.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                30.09.2018
# Last Modified Date:  05.01.2022
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

from collections.abc import KeysView
from pydantic import BaseModel, BaseConfig, Extra

class AmpelBaseModel(BaseModel):
	""" Raises validation errors if extra fields are present """

	class Config(BaseConfig):
		arbitrary_types_allowed = True
		underscore_attrs_are_private = True
		allow_population_by_field_name = True
		validate_all = True
		extra = Extra.forbid

	def __init__(self, **kwargs) -> None:
		self.__config__.extra = Extra.forbid
		try:
			super().__init__(**kwargs)
		except Exception as e:
			raise TypeError(e) from None
		self.__config__.extra = Extra.allow


	@classmethod
	def get_model_keys(cls) -> KeysView[str]:
		return cls.__fields__.keys()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-interface/ampel/abstract/AbsT1ComputeUnit.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                13.05.2021
# Last Modified Date:  13.05.2021
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

from typing import Tuple
from ampel.types import UBson, StockId
from ampel.base.AmpelABC import AmpelABC
from ampel.base.decorator import abstractmethod
from ampel.base.LogicalUnit import LogicalUnit
from ampel.content.DataPoint import DataPoint
from ampel.struct.UnitResult import UnitResult
	

class AbsT1ComputeUnit(AmpelABC, LogicalUnit, abstract=True):

	@abstractmethod
	def compute(self, datapoints: list[DataPoint]) -> tuple[UBson | UnitResult, StockId]:
		...



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-interface/ampel/abstract/AbsTiedT2Unit.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                06.02.2021
# Last Modified Date:  31.05.2021
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

from typing import Sequence
from ampel.model.UnitModel import UnitModel
from ampel.base.AmpelABC import AmpelABC
from ampel.base.LogicalUnit import LogicalUnit


class AbsTiedT2Unit(AmpelABC, LogicalUnit, abstract=True):
	"""
	A T2 unit that depends on the results of other T2 units.
	"""

	#: Dependencies configuration for the underlying tied t2 unit
	#: Sub-class can add unit id constraints by parametrizing UnitModel
	t2_dependency: Sequence[UnitModel]

	def __init__(self, **kwargs):
		"""
		Dependency errors should not happen with config built by ampel (ConfigBuilder)
		but might happen in developers notebook
		"""

		if isinstance(kwargs.get("t2_dependency"), dict):
			kwargs["t2_dependency"] = [kwargs["t2_dependency"]]

		super().__init__(**kwargs)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-interface/ampel/view/T3DocView.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                01.12.2021
# Last Modified Date:  12.12.2021
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

from datetime import datetime
from typing import Any, Literal, overload
from collections.abc import Sequence
from ampel.types import StockId, UBson, TBson, Tag
from ampel.content.MetaRecord import MetaRecord
from ampel.content.T3Document import T3Document
from ampel.config.AmpelConfig import AmpelConfig


class T3DocView:
	"""
	View of a given T3Document.
	A t3 view contains read-only information from a T3Document
	and provides convenience methods to access it.
	"""

	__slots__ = 'unit', 'confid', 'config', 'stock', 'tag', 'code', 'meta', 'body'

	stock: None | Sequence[StockId]
	unit: str
	confid: int
	config: None | dict[str, Any]
	tag: None | Tag | Sequence[Tag]
	code: int
	meta: MetaRecord
	body: UBson


	@classmethod # Static ctor
	def of(cls, doc: T3Document, conf: AmpelConfig) -> "T3DocView":

		if 'config' in doc:
			config = doc['config']
		else:
			config = conf._config['confid'][doc['confid']]

		return cls(
			unit = doc['unit'],
			tag = doc.get('tag') or [],
			code = doc['code'],
			meta = doc['meta'],
			stock = doc.get('stock'),
			body = doc.get('body'),
			config = config,
			confid = doc['confid']
		)


	def __init__(self,
		unit: str,
		code: int,
		meta: MetaRecord,
		confid: int,
		tag: None | Tag | Sequence[Tag] = None,
		stock: None | Sequence[StockId] = None,
		config: None | dict[str, Any] = None,
		body: UBson = None
	):
		sa = object.__setattr__
		sa(self, 'stock', stock)
		sa(self, 'unit', unit)
		sa(self, 'tag', tag)
		sa(self, 'code', code)
		sa(self, 'body', body)
		sa(self, 'meta', meta)
		sa(self, 'config', config)
		sa(self, 'confid', confid)


	def __setattr__(self, k, v):
		raise ValueError("T3DocView is read only")


	def __delattr__(self, k):
		raise ValueError("T3DocView is read only")


	def serialize(self) -> dict[str, Any]:
		return {k: getattr(self, k) for k in self.__slots__ if k != '_frozen'}


	@overload
	def get_body(self) -> None | dict[str, Any]:
		...
	@overload
	def get_body(self, *, raise_exc: Literal[True]) -> dict[str, Any]:
		...
	@overload
	def get_body(self, ret_type: type[TBson]) -> None | TBson:
		...
	@overload
	def get_body(self, ret_type: type[TBson], *, raise_exc: Literal[True]) -> TBson:
		...
	@overload
	def get_body(self, ret_type: type[TBson], *, raise_exc: Literal[False]) -> None | TBson:
		...
	def get_body(self, ret_type: type[TBson] = dict, *, raise_exc: bool = False) -> None | TBson: # type: ignore[assignment]
		"""
		:param raise_exc: raise exception if the body has not the expected type

		reveal_type(t3dv.get_body()) -> Union[builtins.dict[builtins.str, Any], None]
		reveal_type(t3dv.get_body(raise_exc=True)) -> builtins.dict[builtins.str, Any]
		reveal_type(t3dv.get_body(str)) -> Union[builtins.str, None]
		reveal_type(t3dv.get_body(str, raise_exc=True)) -> builtins.str
		"""
		return self.body if isinstance(self.body, ret_type) else None


	@overload
	def get_time_created(self, to_string: Literal[False]) -> None | float:
		...
	@overload
	def get_time_created(self, to_string: Literal[True]) -> None | str:
		...
	def get_time_created(self, to_string: bool = False) -> None | float | str:

		if to_string:
			return datetime.fromtimestamp(self.meta['ts']).strftime('%d/%m/%Y %H:%M:%S')

		return self.meta['ts']



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-interface/ampel/secret/NamedSecret.py
# License:             BSD-3-Clause
# Author:              Jakob van Santen <jakob.van.santen@desy.de>
# Date:                14.08.2020
# Last Modified Date:  30.12.2021
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

from ampel.secret.Secret import Secret, T
from ampel.base.AmpelBaseModel import AmpelBaseModel


class NamedSecret(Secret[T], AmpelBaseModel):
	"""
	A Secret:
	- featuring a label used as lookup key during secret resolution
	- holding a simple reference to a sensitive payload value
	"""

	label: str
	value: None | T = None

	def __repr__(self):
		return '<NamedSecret>'

	def get(self) -> T:
		if self.value is None:
			raise ValueError("Secret not yet resolved")
		return self.value

	def set(self, arg: T) -> None:
		if self.value:
			raise ValueError("Secret already resolved")
		self.value = arg



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-interface/ampel/model/operator/OneOf.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                22.10.2018
# Last Modified Date:  18.03.2021
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

from typing import Generic
from ampel.types import T
from ampel.base.AmpelGenericModel import AmpelGenericModel

class OneOf(AmpelGenericModel, Generic[T]):
	one_of: list[T]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-interface/ampel/content/StockDocument.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                28.12.2019
# Last Modified Date:  05.05.2021
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

from typing import Any, TypedDict, Literal
from collections.abc import Sequence
from ampel.types import StockId, ChannelId, Tag
from ampel.content.JournalRecord import JournalRecord


class StockDocument(TypedDict, total=False):
	"""
	The stock record ties together data from various sources, selected by
	various channels, but all related to the same underlying object. Each
	channel has a different view of the stock.
	From the perspective of a given channel, the stock is updated
	whenever a linked document (T0, T1, T2) is updated.

	A dict containing 1 or more of the following items:
	"""

	#: The unique id associated with the stock. Integer most of the time
	stock: StockId

	#: Optional source origin (avoids potential stock collision between different data sources)
	origin: int

	#: Optional tag(s)
	tag: Sequence[Tag]

	#: Channels asscoiated with this stock
	channel: Sequence[ChannelId]

	#: Records of activity
	journal: Sequence[JournalRecord]

	#: Creation time (UNIX epoch) in each channel
	ts: dict[ChannelId, dict[Literal['tied', 'upd'], float]]

	#: Last update time for any channel
	updated: int | float

	#: External name(s) associated with the stock
	name: Sequence[int | str]

	#: Optional specific content
	body: dict[str, Any]



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['ampel',
 'ampel.abstract',
 'ampel.alert',
 'ampel.base',
 'ampel.cli',
 'ampel.config',
 'ampel.content',
 'ampel.enum',
 'ampel.model',
 'ampel.model.operator',
 'ampel.protocol',
 'ampel.secret',
 'ampel.struct',
 'ampel.util',
 'ampel.view']

package_data = \
{'': ['*']}

install_requires = \
['PyYAML>=6.0.0,<7.0.0',
 'pydantic>=1.9.0,<2.0.0',
 'ujson>=5.1.0,<6.0.0',
 'xxhash>=3.0.0,<4.0.0']

extras_require = \
{'docs': ['Sphinx>=4.0.0,<5.0.0',
          'sphinx-autodoc-typehints>=1.11.1,<2.0.0',
          'tomlkit>=0.11.0,<0.12.0']}

setup_kwargs = {
    'name': 'ampel-interface',
    'version': '0.8.3a14',
    'description': 'Base classes for the Ampel analysis platform',
    'long_description': '# Ampel-interface\n\n`ampel-interface` provides type-hinted abstract base classes for [Ampel](https://ampelproject.github.io).',
    'author': 'Valery Brinnel',
    'author_email': None,
    'maintainer': 'Jakob van Santen',
    'maintainer_email': 'jakob.van.santen@desy.de',
    'url': 'https://ampelproject.github.io',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-interface/ampel/content/T2Document.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                13.01.2018
# Last Modified Date:  25.06.2022
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

from typing import TypedDict
from typing_extensions import Required
from collections.abc import Sequence
from ampel.types import UBson, ChannelId, StockId, Tag, UnitId, T2Link
from ampel.content.MetaRecord import MetaRecord


class T2Document(TypedDict, total=False):
	"""
	Specifications for tier2 documents stored as BSON structures in the ampel DB.
	Calculations of the associated t2 unit is performed based on ampel data referenced by the attribute 'link'.
	Linked input data type can be either :class:`~ampel.content.StockDocument.StockDocument`,
	:class:`~ampel.content.DataPoint.DataPoint`, or :class:`~ampel.content.T1Document.T1Document`.
	"""

	#: Stock id associated with the data
	stock: Required[StockId | Sequence[StockId]]

	#: Optional source origin (avoids potential stock collision between different data sources)
	origin: int

	#: Name of the unit to be run. This may be hashed for performance reasons.
	unit: Required[UnitId]

	#: Configuration hash, if unit defaults were overridden. The underlying values can be resolved with
	#: :meth:`UnitLoader.get_init_config() <ampel.core.UnitLoader.UnitLoader.get_init_config>`
	config: Required[None | int]

	#: References to input data
	link: Required[T2Link]

	#: visible by any projection (not channel bound)
	tag: Sequence[Tag]

	#: Ampel channel(s) associated with this document
	channel: Required[Sequence[ChannelId]]

	#: Records of activity on this document
	meta: Required[Sequence[MetaRecord]]

	#: Name of the database collection holding the input data (t1 if unspecified)
	#: (enables efficient DB queries at T3 level)
	col: str

	#: DocumentCode.NEW for new T2 document, DocumentCode.OK if computation was successful
	code: Required[int]

	#: value(s) returned by T2 unit execution(s)
	body: Required[Sequence[UBson]]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-interface/ampel/model/operator/AnyOf.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                15.10.2018
# Last Modified Date:  18.03.2021
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

import collections
from typing import Generic
from ampel.types import T
from ampel.model.operator.AllOf import AllOf
from ampel.base.AmpelGenericModel import AmpelGenericModel


class AnyOf(AmpelGenericModel, Generic[T]):

	#: Select items by logical OR
	any_of: list[T | AllOf[T]]

	def __init__(self, **kwargs) -> None:
		if 'any_of' in kwargs and not isinstance(kwargs['any_of'], collections.abc.Sequence):
			kwargs['any_of'] = [kwargs['any_of']]
		super().__init__(**kwargs)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-interface/ampel/cli/AmpelArgumentParser.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                17.03.2021
# Last Modified Date:  27.06.2022
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

import textwrap
from os import environ
from typing import Any
from ampel.cli.MaybeIntAction import MaybeIntAction
from ampel.cli.LoadJSONAction import LoadJSONAction
from ampel.cli.LoadAnyOfAction import LoadAnyOfAction
from ampel.cli.LoadAllOfAction import LoadAllOfAction
from argparse import ArgumentParser, _ArgumentGroup
from ampel.cli.AmpelHelpFormatter import AmpelHelpFormatter


class AmpelArgumentParser(ArgumentParser):
	"""
	Argument parser with additional features.
	Supports adding notes and examples.

	Some differences with ArgumentParser:
	- The "required argument" group is placed first
	- The "-h --help" line is not printed
	- Every section is capitalized
	- Argument values are represented by hash rather than capitalized argument names:
		* ArgumentParser: -parameter PARAMETER [PARAMETER ...]
		* This:           -parameter # [# ...]
	- New representations were added:
		* Arguments requiring at least two values: -parameter # # ...
		* Arguments requiring a JSON string: -parameter '{#}'
	- A "notation section" is generated dynamically and added automatically
	- The arguments column (left) is larger (width of 40 vs 27 for ArgumentParser's default formatter)

	Note:
	The parameter -config (path to an ampel config file) is automatically moved from the
	*required* argument group to the *optional* argument group if the environment variable
	AMPELCONF exists and points to a valid file. In this case, a note is added and
	"-config ampel_conf.yaml" is stripped out of added examples automatically (deactivable)

	Conveniences methods:
	- move_group_up: moves a group up in order
	- create_logic_args: creates 4 arguments for a given name.
	  For example, parser.create_opt_logic_args("channel", "Channel") will create:
	  -channel #                 Channel to be matched (non-exclusively)
	  -channels-or # # ...       Channels to be matched (OR connected)
	  -channels-and # # ...      Channels to be matched (AND connected)
	  -channels-json '{#}'       Channels to be matched (JSON, see note below)
	  whereby a note regarding parameters with suffix *-json is added automatically.
	  The values of channels-or, channels-and and channels-json have 'channel' as destination
	  so that the calling code should not have to differenciate between these.
	  Moreover, each generated argument is associated with a dedicated action:
	  MaybeIntAction, LoadAnyOfAction, LoadAllOfAction and LoadJSONAction in ampel/cli

	See the class ampel.cli.T2Command for an full utilization of this class
	"""

	def __init__(self, ampel_op: None | str = None, **kwargs) -> None:

		super().__init__(formatter_class=AmpelHelpFormatter, **kwargs)
		self.notes: list[str] = []
		self.examples: list[str] = []
		self.args_descr: dict[str, str] = {}
		self._logic_ops_note_added = False
		self.notations: set[str | tuple[str, str]] = set()
		self.has_env_conf = False
		self.groups: dict[str, _ArgumentGroup] = {}
		self.bullet = "\u2022"
		self.usage = "" # required by argparse (AssertionError), unclear why
		self._ampel_op = ampel_op

		self.spacer = "  "
		self.note_open = "["
		self.note_close = "]"
		self.ex_open = "|"
		self.ex_close = "|"


		# Pop optional group to reorder it, putting required arguments first (really questionable defaults there...)
		optional_group = self._action_groups.pop()

		# Add a reference by name
		self.groups["optional"] = optional_group

		# Suppress "-h, --help show this help message and exit" (chicken and egg)
		# optional._option_string_actions = {}
		optional_group._group_actions = []

		# Capitalize title
		optional_group.title = optional_group.title.capitalize() # type: ignore

		# Add required argument group (first pos)
		self.groups["required"] = self.add_argument_group('Required arguments')

		# Re-insert optional group
		self._action_groups.append(optional_group)

		# TODO: use better regex matching
		if "AMPEL_CLI_OPTS" in environ and "config " in environ["AMPEL_CLI_OPTS"]:
			self.has_env_conf = True


	def set_ampel_sub_op(self, name: str) -> None:
		self.ampel_sub_op = name
	

	def set_bullet(self, bullet: str) -> None:
		""" Used for notes and examples. Bots might require markdown """
		self.bullet = bullet


	def add_group(self, name: str, title: str, **kwargs):
		self.groups[name] = self.add_argument_group(name, **kwargs)
		self.groups[name].title = title


	def set_group_defaults(self, name: str, **kwargs):
		self.groups[name].set_defaults(**kwargs)


	def add_description(self, descr: None | str | list[str]) -> None:

		if not descr:
			return

		if isinstance(descr, str):
			descr = [
				ell for el in descr.split("\n")
				for ell in textwrap.wrap(el, width=100)
			]

		if self.description is None:
			self.description = self.spacer

		self.description += f"\n{self.spacer}".join(descr)


	def move_group_up(self, group: _ArgumentGroup):
		if (p := self.get_group_index_pos(group)) > 0:
			self._action_groups.remove(group)
			self._action_groups.insert(p - 1, group)


	def get_group_index_pos(self, group: _ArgumentGroup) -> int:
		for i, el in enumerate(self._action_groups):
			if el == group:
				return i
		return -1


	def get_group(self, name: str) -> None | _ArgumentGroup:
		return self.groups.get(name)


	def add_arg(self,
		name: str, group: str | _ArgumentGroup = "optional",
		help: None | str = None, **kwargs
	) -> None:
		"""
		:param target:
		Example:
			- add_arg("verbose", "optional")
			- add_arg("out", "required")
		"""

		self._auto_metavar(kwargs)

		# Add only the nomemclature elements that are actually used
		if "const" in kwargs:
			self.notations.add(("-argument [#]", "Argument with an overridable default value"))
		elif kwargs.get("nargs") == "+":
			self.notations.add(("-argument # [# ...]", "Argument accepting one or more values"))
		elif isinstance(kwargs.get("action"), str):
			self.notations.add(("-argument", "Argument without value (flag)"))
		else:
			self.notations.add(("-argument #", "Argument requiring one value"))

		if name == "config" and group == "required" and self.has_env_conf:
			self.add_note(
				"Config file defined by env var AMPEL_CLI_OPTS will be used " +
				"unless an override using -config is specified"
			)
			group = "optional"

		if isinstance(group, str):
			group = self.groups[group]

		group.add_argument(
			f"--{name}", help=help or self.args_descr.get(name),
			required=True if group == self.groups["required"] else False, **kwargs
		)


	def add_x_args(self, group: str = "optional", *args: dict[str, Any]):
		"""
		Create a mutually exclusive group, attach it to the group defined in 'target'
		and add arguments to it according to provided definitions (*args)
		"""
		self.notation_add_mutual_exclusivity()

		try:
			# Add newline between mutually excl groups
			if (
				self.groups[group]._mutually_exclusive_groups[-1]._actions[-1].dest ==
				self.groups[group]._actions[-1].dest
			):
				last_args = self.groups[group]._option_string_actions[
					list(self._option_string_actions.keys())[-1]
				]
				last_args.help = (last_args.help or "") + "\n\n"
		except Exception:
			pass
		x = self.groups[group].add_mutually_exclusive_group()
		for arg in args:
			self.add_arg(group=x, **arg)


	def _auto_metavar(self, kw: dict) -> None:
		if "dest" in kw or "metavar" in kw:
			return
		if (a := kw.get("action")) and a != MaybeIntAction:
			return
		kw['metavar'] = "#"


	def set_help_descr(self, args_descr: dict) -> None:
		""" Sets help parameter descriptions """
		self.args_descr = args_descr


	def add_note(self, note: str, pos: None | int = None, ref: None | str = None) -> None:
		""" Newlines in note are supported and will be properly formatted """

		notes = note.split("\n")
		for i in range(len(notes)):
			if len(notes[i]) > 100:
				notes[i:i+1] = textwrap.wrap(notes[i], width=100)

		start = f"{self.spacer}{self.bullet} "
		content = start

		if ref:
			if ref == "#":
				raise ValueError("Hash sign cannot be used as ref")
			content += f"{self.note_open}{ref}{self.note_close} "

		for i, el in enumerate(notes):
			if i > 0:
				content += "\n" + " " * len(start)
			content += el

		if pos is not None:
			self.notes.insert(pos, content)
		else:
			self.notes.append(content)


	def add_notes(self, notes: list[str]) -> None:
		""" Each note will start with a bullet """
		for el in notes:
			self.add_note(el)


	def add_example(self,
		ex: str, prepend="ampel ", append="", auto_strip_config: bool = True,
		ref: None | str = None
	) -> None:
		if ref:
			prepend = f"{self.ex_open}{ref}{self.ex_close} {prepend}"
		payload = f"{self.spacer}{self.bullet} {prepend}{ex}{append}"
		if auto_strip_config and self.has_env_conf:
			self.examples.append(payload.replace(" -config ampel_conf.yaml ", " "))
		else:
			self.examples.append(payload)


	def create_logic_args(self,
		name: str, group: str, descr: str, metavar: str = "#",
		required: bool = False, pos: None | int = None,
		ref: None | str = None, excl: bool = False,
		json: bool = True, **kwargs
	) -> None:
		"""
		Ex: channel, channels-and, channels-or, channels-json
		Ex: with-tag, with-tags-and, with-tags-or, with-tags-json
		"""
		
		what = "excluded" if excl else "matched"
		dest = name.replace("-", "_")

		# Cosmetic
		if not self._logic_ops_note_added:
			last_args = self.groups[group]._option_string_actions[
				list(self._option_string_actions.keys())[-1]
			]
			last_args.help = (last_args.help or "") + "\n\n"

		self.notation_add_mutual_exclusivity()
		self.notations.add(("-argument # # ...", "Argument requiring at least two values"))
		
		mux = self.groups[group].add_mutually_exclusive_group(required=required)
		mux.add_argument(
			f"--{name}", metavar=metavar,
			type=str, nargs=1, default=None, action=MaybeIntAction,
			help=f"{descr} to be {what} (non-exclusively)"
		)

		mux.add_argument(
			f"--{name}s-or", dest=dest, action=LoadAnyOfAction,
			metavar=metavar, type=str, default=None, nargs="+",
			help=f"{descr}s to be {what} (OR connected)"
		)

		mux.add_argument(
			f"--{name}s-and", dest=dest, action=LoadAllOfAction,
			metavar=metavar, type=str, default=None, nargs="+",
			help=f"{descr}s to be {what} (AND connected)" + ("" if json else "\n\n")
		)

		if json:
			suffix = f"{self.note_open}{ref}{self.note_close}" if ref else ""
			mux.add_argument(
				f"--{name}s-json", dest=dest, action=LoadJSONAction,
				metavar=metavar, type=str, default=None,
				help=f"{descr}s to be {what} {suffix}\n\n"
			)

			# Add note about JSON arg
			if not self._logic_ops_note_added:
				self.add_note(
					"Allows the use of logic operators such as:\n" +
					"Nested logic: \'{\"any_of\": [\"VAL1\", {\"all_of\": [\"VAL2\", \"VAL3\"]}]}\'\n" +
					"Exclusive match: \'{\"one_of\": [\"VAL1\"]}\'", pos=pos, ref=ref
				)
				self._logic_ops_note_added = True


	# Cosmetic
	def error(self, message: str) -> Any:
		""" As for now, there is no better way to do this  """
		if "-" in message:
			message = message.replace(" --", " -")
		if message.startswith("the following arguments are required:"):
			if len(message.split("-")) == 2:
				return super().error(message.replace("arguments are", "argument is"))
		return super().error(message)


	def notation_add_note_references(self):
		self.notations.add(f"{self.spacer}Note references are marked with [ ]")


	def notation_add_example_references(self):
		self.notations.add(f"{self.spacer}Example references are marked with | |")


	def notation_add_mutual_exclusivity(self):
		self.notations.add(
			f"{self.spacer}Consecutive mutually exclusive arguments are marked with \u22BB"
		)
		

	def hint_query_logic(self, pos: None | int = None, ref: None | str = None):
		self.add_note(
			"Matching criteria related to different keys are AND-combined with each other",
			pos, ref
		)


	def hint_time_format(self, pos: None | int = None, ref: None | str = None):
		self.add_note(
			"Date-time strings are parsed using datetime.fromisoformat(<value>).\n" +
			"Example of supported formats (unexhaustive): 2011-11-04 or 2011-11-04T00:05:23",
			pos, ref
		)


	def hint_config_override(self, pos: None | int = None, ref: None | str = None):
		self.add_note(
			"Any existing config parameter can be overriden using -path.to.config.key value\n" +
			"Example: -mongo.prefix AmpelTest",
			pos, ref
		)


	def print_help(self,  # type: ignore[override]
		show_description: bool = True,
		show_usage: bool = True,
		show_notation: bool = True,
		show_arguments: bool = True,
		show_notes: bool = True,
		show_examples: bool = True
	):

		if show_description and self.description:
			print("\nOperation:\n" + self.description)

		if show_usage:
			print("\nUsage:")
			sub_op = getattr(self, "ampel_sub_op", None) # ex: show
			if self._ampel_op:
				if sub_op: # "ampel log show"
					print(f"{self.spacer}ampel {self._ampel_op} {sub_op} <arguments>")
				else: # "ampel run" (No sub-op/action defined)
					print(f"{self.spacer}ampel {self._ampel_op} <arguments>")
			else: # "ampel"
				print(f"{self.spacer}ampel <operation> <arguments>")

		# Core (arguments) help section
		formatter = self._get_formatter()
		for action_group in self._action_groups:
			formatter.start_section(action_group.title)
			formatter.add_text(action_group.description)
			formatter.add_arguments(action_group._group_actions)
			formatter.end_section()

		fh = formatter.format_help()

		# Cosmetic: use numbers for references
		fh = self._numerate(fh, self.notes, self.note_open, self.note_close)
		fh = self._numerate(fh, self.examples, self.ex_open, self.ex_open)

		if show_notation and self.notations:
			cw = getattr(formatter, "_action_max_length", 30) # col width
			print("\nNotation:")
			if (x := sorted([v for v in self.notations if isinstance(v, str)], key=lambda x: len(x))):
				print("\n".join(x) + "\n")
			raw_not = sorted(
				[v for v in self.notations if isinstance(v, tuple)],
				key=lambda x: len(x[0])
			)
			print("\n".join([self._format(v[0], v[1], cw) for v in raw_not]))

		if show_arguments:
			print("\n" + fh, end="")

		if show_notes and self.notes:
			# Put numerated notes first
			print("\nNote%s:" % ("s" if len(self.notes) > 1 else ""))
			print("\n".join(self._sort(self.notes)))

		if show_examples and self.examples:
			print("\nExample%s:" % ("s" if len(self.examples) > 1 else ""))
			print("\n".join(self._sort(self.examples)) + "\n")

		if self.epilog:
			print(self.epilog.strip())

		print("")


	def _format(self, k: str, v: str, l: int) -> str:
		filler = (l - len(k) - 1) * " "
		return self.spacer + k + filler + v


	def _numerate(self, fh: str, target: list[Any], open_key: str, close_key: str) -> str:
		try:
			l: list[str] = []
			y = 1
			for i, el in enumerate(target):
				if isinstance(el, str) and el[4] == open_key:
					l.append(el[5])
					target[i] = el[:5] + str(y) + el[6:]
					y += 1
			for i, el in enumerate(l, 1):
				fh = fh.replace(
					open_key + el + close_key,
					open_key + str(i) + close_key
				)
		except Exception:
			pass

		return fh

	def _sort(self, target: list[Any]) -> list[str]:
		""" # Put numerated notes/examples first. If not notes, then examples """
		opn = self.note_open if target == self.notes else self.ex_open
		return [el for el in target if el[4] == opn] + \
			[el for el in target if el[4] != opn]


	@classmethod
	def build_choice_help(cls,
		op: str, sub_ops: list[str], hlp: dict[str, str],
		description: None | str = None
	) -> 'AmpelArgumentParser':

		parser = AmpelArgumentParser(ampel_op=op)
		parser.set_ampel_sub_op("<action>")

		if description:
			parser.add_description(description)

		sp_action = parser.add_subparsers(parser_class=cls)
		parser._action_groups[0].title = "Actions"

		for s in sub_ops:
			sp_action.add_parser(s, help=hlp.get(s))

		return parser



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-interface/ampel/util/recursion.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                16.11.2021
# Last Modified Date:  16.11.2021
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

from collections.abc import Callable


def walk_and_process_dict(
	arg: dict | list, callback: Callable,
	match: None | list[str] = None, path: str = None, **kwargs
) -> bool:
	"""
	callback is called with 4 arguments:
	1) the path of the possibly nested entry. Ex: 'processor.config.select' or 'processor'
	2) the matching key (from list 'match'). Ex: 'config'
	3) the matching (sub) dict
	4) the **kwargs provided to this method
	and should return True if a modification was performed

	Simplest callback function:
	def my_callback(path, k, d):
		print(f'{path} -> {k}: {d}\n')
		return False

	:returns: True if a modification was performed, False otherwise
	"""

	ret = False

	if isinstance(arg, list):
		for i, el in enumerate(arg):
			ret = walk_and_process_dict(
				el, callback, match, f'{path}.{i}' if path else f'{i}', **kwargs
			) or ret

	if isinstance(arg, dict):

		for k, v in arg.items():

			if not match or k in match:
				ret = callback(path, k, arg, **kwargs) or ret

			if isinstance(v, dict):
				ret = walk_and_process_dict(
					v, callback, match, f'{path}.{k}' if path else f'{k}', **kwargs
				) or ret

			if isinstance(v, list):
				for i, el in enumerate(v):
					ret = walk_and_process_dict(
						el, callback, match, f'{path}.{k}.{i}' if path else f'{k}.{i}', **kwargs
					) or ret

	return ret



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-interface/ampel/content/T1Document.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                Unspecified
# Last Modified Date:  25.06.2022
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

from typing import Literal, TypedDict
from typing_extensions import Required
from collections.abc import Sequence
from ampel.types import UBson, StockId, DataPointId, ChannelId, Tag, UnitId
from ampel.content.MetaRecord import MetaRecord


class T1Document(TypedDict, total=False):
	"""
	A symbolic collection of :class:`~ampel.content.DataPoint.DataPoint`,
	representing the state of a stock at a given point in time as viewed
	through one or more channels.

	Dict containing 1 or more of the following items:
	"""

	#: Name of AbsT1ComputeUnit potentially associated with this doc
	unit: UnitId

	#: Optional hashed config of t1 unit
	config: int

	#: stock(s) this doc is associated to
	stock: Required[StockId | Sequence[StockId]]

	#: Optional instrument/source identifier
	origin: int

	#: Result of AbsT1CombineUnit units
	dps: Required[Sequence[DataPointId]]

	#: intger hash hash of dps (referenced by stated T2 units)
	link: Required[int]

	#: visible by any projection (not channel bound)
	tag: Sequence[Tag]

	#: Ampel channel(s) associated with this document
	channel: Required[Sequence[ChannelId]]

	#: DocumentCode.NEW for T1 units requiring computation, DocumentCode.OK otherwise
	code: Required[int]

	#: References among other things the id of process invocation that created this doc
	meta: Required[Sequence[MetaRecord]]

	#: Potential result of AbsT1ComputeUnit subclasses
	body: Sequence[UBson]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                ampel/utils/docstringutils.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                02.09.2018
# Last Modified Date:  11.12.2019
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

import inspect, re

regex_ignore = re.compile(':[ \t]*?ClassVar|[ \t]*?#')
regex_stop = re.compile('^[ \t]*?@|^[ \t]*?def|^class ')

def gendocstring(klass):
	"""
	=============================================================================
	Decorator for model sub-classes and python 3.7 dataclasses.
	-> Automatically generates doctring based on class members (makes required
	variables (including type hints) available in docstring)
	=============================================================================

	Note: the module 'inspect' does not work with classes defined within ipython
	
	Code example:
	~~~~~~~~~~~~~
	
		from ampel.base.AmpelBaseModel import AmpelBaseModel
		from ampel.util.docstringutils import gendocstring
	
		@gendocstring
		class MyConfig(AmpelBaseModel):
			my_str: str
			my_int: int = 0
	
		@gendocstring
		class MyConfig2(AmpelBaseModel):
			\"\"\"
			Existing docstrings will be preserved
			\"\"\"
			my_str2: str
			my_int2: int = 0

	Generated docstrings:
	~~~~~~~~~~~~~~~~~~~~~
	
		In []: print(MyConfig.__doc__)
		Out []:
		===================
		Fields:
		  my_str: str
		  my_int: int = 0
		===================
	
		In []: print(MyConfig2.__doc__)
		Out []:
		=====================================
		Existing docstrings will be preserved
		-------------------------------------
		Fields:
		  my_str2: str
		  my_int2: int = 0
		=====================================
	"""
	out_doc = []
	exisiting_doc = []
	in_doc = inspect.getdoc(klass)

	if in_doc:
		# Gather existing docstring
		for el in in_doc.split('\n'):
			sel = el.strip()
			if not sel:
				continue
			exisiting_doc.append(el)

	for el in inspect.getsource(klass).split('\n')[1:]:

		ell = el.strip()

		# Ignore non field-definitions
		if not ell or ":" not in ell or regex_ignore.findall(ell):
			continue

		if regex_stop.findall(ell):
			break
			
		out_doc.append("  " + ell)

	max_len = max([len(el) for el in out_doc+exisiting_doc]+[19])
	klass.__doc__ = "="*max_len + "\n"

	if exisiting_doc:
		klass.__doc__ += "\n".join(exisiting_doc) + "\n" + "-"*max_len + "\n"

	klass.__doc__ += \
		"Fields: \n" + \
		"\n".join(out_doc) + \
		"\n" + "="*max_len + "\n"
	
	return klass



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-interface/ampel/secret/AESecret.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                20.06.2021
# Last Modified Date:  30.12.2021
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

from ampel.secret.Secret import Secret
from ampel.base.AmpelBaseModel import AmpelBaseModel


class AESecret(Secret[str], AmpelBaseModel):
	"""
	AES encrypted secret.

	To create an encrypted secret, please:
	- go to https://bitwiseshiftleft.github.io/sjcl/demo/
	- enter the shared password in the green box
	- enter the secret message (authtoken for example) in the red box
	- leave authenticated data empty
	- in "Cipher Parameters", check the option CCM. (OCB2 will *not* work)
	- click on the red arrow "encrypt"
	- copy the "Ciphertext" JSON dict
	- use it in your config (section 'secrets')

	Make sure the 'shared password' used in step 2 is known to us.
	"""

	iv: str
	v: int
	iter: int
	ks: int
	ts: int
	mode: str
	adata: str
	cipher: str
	salt: str
	ct: str

	value: None | str

	def __repr__(self):
		return '<AESecret>'

	def set(self, value: str) -> None:
		self.value = value

	def get(self) -> str:
		if self.value is None:
			raise ValueError("AES payload not decrypted yet")
		return self.value



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-interface/ampel/util/bson.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                24.05.2020
# Last Modified Date:  24.05.2020
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

from bson.binary import Binary

def int_to_bindata(int_arg: int) -> Binary:
	"""
	converts a python integer number (unlimited length) into a BSON data type 'BinData'.
	The used subtype 0 (\\x00): "Generic binary subtype"
	The int to bytes conversion uses the little Endian byte ordering
	(most significant byte is at the end of the byte array)
	"""
	return Binary(
		int_arg.to_bytes(
			(int_arg.bit_length() + 7) // 8,
			'little'
		),
		0
	)

def bindata_to_int(bin_data_bytes: Binary) -> int:
	"""
	converts a BSON data type 'BinData' (subtype 0) into a python integer number
	The little Endian byte ordering is used
	(most significant byte is at the end of the byte array)
	"""
	return int.from_bytes(bin_data_bytes, byteorder='little')



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-interface/ampel/base/LogicalUnit.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                12.10.2019
# Last Modified Date:  09.01.2022
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

from typing import Any, ClassVar
from ampel.types import Traceless
from ampel.base.AmpelUnit import AmpelUnit
from ampel.protocol.LoggerProtocol import LoggerProtocol


class LogicalUnit(AmpelUnit):
	"""
	Logical as in: performing logic operations of ampel's tier: add, combine, augment, synthetise.
	Base for standardized t0, t1, t2 and t3 units.
	Note:
	- the defined parameters must be serializable
	- logical units are initialized (by UnitLoader) as follows:
	* ctor
	* if available, secrets are resolved
	* if defined by sub-class, post_init() is called
	"""

	logger: Traceless[LoggerProtocol]

	#: Resources requirements as class variable (passed on to and merged with subclasses).
	require: ClassVar[None | tuple[str, ...]] = None

	resource: Traceless[None | dict[str, Any]] = None

	#: Private variable potentially set by UnitLoader for provenance purposes. Either:
	#: * None if provanance flag is False
	#: * 0 in case model content is not serializable
	#: * any other signed int value
	_trace_id: None | int = None

	# Some unit contributors might want to restrict units usage
	# by scoping them to their respective distribution name (str)
	# private: ClassVar[None | str] = None


	@classmethod
	def __init_subclass__(cls, **kwargs) -> None:
		super().__init_subclass__(**kwargs) # type: ignore
		setattr(cls, "require",
			set(getattr(cls, "require", None) or []) | {
				el for base in cls.__bases__
				for el in (getattr(base, 'require', None) or [])
			}
		)
